import sys
import argparse
from pathlib import Path
from .core import print_startup_message
from .tuning import main as tuning_main

def main():
    parser = argparse.ArgumentParser(print_startup_message())
    parser.add_argument("input_file", help="Input file (e.g., tuning.in)")
    
    args = parser.parse_args()
    
    input_path = Path(args.input_file)
    if not input_path.exists():
        print(f"Error: Input file '{input_path}' not found")
        sys.exit(1)
    
    # Execute tuning module with the input file
    tuning_main(str(input_path))

if __name__ == "__main__":
    main()


import numpy as np
import os,subprocess,sys
from pyscf import gto, scf
from pyscf.solvent import pcm,smd
from pyscf.geomopt.berny_solver import optimize

def create_dft_molecule(atom_input, basis_set, functional='m06-2x', charge=0, spin=0):
    """
    Create a PySCF molecule and KS/DFT object from either:
    - atom_input: list of (atom_symbol, (x,y,z)) tuples
    - atom_input: string path to XYZ file
    """
    mol = gto.Mole(basis = basis_set, charge=charge, spin=spin, xc=functional)
    if isinstance(atom_input, str) and os.path.isfile(atom_input):
        mol.atom = atom_input 
    else:
        mol.atom = atom_input  
    mf = scf.UKS(mol)
    return mf


def optimize_and_get_equilibrium(mf):
    """
    Optimize the geometry of the molecule and return the equilibrium geometry.
    """
    mol_eq = optimize(mf)
    coords = mol_eq.atom_coords()
    atoms = [mol_eq.atom_symbol(i) for i in range(mol_eq.natm)]
    atom_list = [(atom, coord) for atom, coord in zip(atoms, coords)]
    return atom_list

#Parameters:
molecule = 'water.xyz'
functional = 'm06-2x'
opt_basis_set = '6-31+G*'
sp_basis_set = 'ccpvdz'
solvent = None

mf = create_dft_molecule(molecule, opt_basis_set,functional=functional)
equilibrium_geometry = optimize_and_get_equilibrium(mf)

if solvent is not None:
    mf = create_dft_molecule(equilibrium_geometry, sp_basis_set,functional=functional)
    mf = mf.PCM()
    mf.with_solvent.eps = smd.solvent_db[solvent][5]
    mf.with_solvent.method = 'C-PCM'
    mf.with_solvent.lebedev_order = 29
else:
    mf = create_dft_molecule(equilibrium_geometry, sp_basis_set)

mf.kernel()


# def get_redox_potential(mf, gfec_basis_set, gfec_functional, sscf_basis_set, sscf_functional, charge, spin, method):
#     #Neutral
#     rdx_n = create_molecule_object(mf.mol.atom, gfec_basis_set, method=method,functional=gfec_functional, charge=charge,spin=spin)
#     rdx_n_opt = optimize_and_get_equilibrium(rdx_n)
#     rdx_n = create_molecule_object(rdx_n_opt,gfec_basis_set, method=method,functional=gfec_functional, charge=charge,spin=spin)
#     _, _, G_corr_rdx_n = get_molecule_gfec(rdx_n)

#     #Anion
#     rdx_a,_,_ = create_charged_molecule_object(mf.mol.atom, gfec_basis_set, method=method,functional=gfec_functional, original_charge=charge, original_spin=spin, charge_change=-1)
#     rdx_a_opt = optimize_and_get_equilibrium(rdx_a)
#     rdx_a,_,_ = create_charged_molecule_object(rdx_a_opt, gfec_basis_set, method=method, functional=gfec_functional, original_charge=charge, original_spin=spin, charge_change=-1)
#     _,_, G_corr_rdx_a = get_molecule_gfec(rdx_a)


#     #hbs = higher basis set SCF
#     rdx_n_hbs = create_molecule_object(rdx_n.mol.atom, sscf_basis_set, functional=sscf_functional, method=method, charge=charge, spin=spin)
#     solvated_n = solvate_molecule(rdx_n_hbs, solvent=solvent)

#     rdx_a_hbs,_,_ = create_charged_molecule_object(rdx_a.mol.atom, sscf_basis_set, functional=sscf_functional, method=method, original_charge=charge, original_spin=spin, charge_change=-1)
#     solvated_a = solvate_molecule(rdx_a_hbs, solvent=solvent)

#     solvated_n.kernel()
#     solvated_a.kernel()

#     #GFE Adjustment
#     gfe_n = solvated_n.e_tot + G_corr_rdx_n
#     gfe_a = solvated_a.e_tot + G_corr_rdx_a

#     #Redox Gibbs Free Energy 
#     dG_red  = -1*((gfe_a - gfe_n) * HARTREE_TO_EV)
#     rdx_pot = (dG_red) -4.43

#     print(f"G_corr_rdx_a: {G_corr_rdx_a}, G_corr_rdx_n: {G_corr_rdx_n}")
#     print(f"gfe_a: {gfe_a}, gfe_n: {gfe_n}, dG_red: {dG_red}, rdx_pot: {rdx_pot}")

#     return rdx_pot


import os, time, subprocess, requests
import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem
from pyscf import gto, scf, dft, tdscf, qmmm, lib
from pyscf.solvent import smd
from pyscf.geomopt.geometric_solver import optimize

# GPU imports
try:
    import cupy as cp
    GPU_AVAILABLE = True
except ImportError:
    cp = None
    GPU_AVAILABLE = False

OFFICE_API = "https://officeapi.akashrajpurohit.com"



##############################################
#             Hardware checks                #
##############################################

def check_gpu_info():
    """
    This function uses CuPy to detect CUDA-capable GPUs on the system.
    """
    if not GPU_AVAILABLE:
        print("\nCuPy not installed - CPU mode only.")
        print("For GPU acceleration: pip install emsuite[gpu]\n")
        return 0
    
    try:
        device_count = cp.cuda.runtime.getDeviceCount()
        if device_count < 1:
            print("\nNo GPUs found.\nSwitching to CPU mode.\n")
            return 0
        else:
            print(f"\n{device_count} GPU(s) detected.\n")
            return device_count
    except Exception as e:
        print(f"\nGPU not available: {e}")
        print("Switching to CPU mode.\n")
        return 0


def check_cpu_info():
    """
    Get the number of available CPU cores on the system.
    
    Returns:
        int: Number of CPU cores available, defaults to 1 if unable to determine
        
    Note:
        Uses os.cpu_count() to detect CPU cores and handles exceptions.
    """
    try:
        cpu_cores = os.cpu_count()
        # print(f"Number of CPU cores available: {cpu_cores}")
        return cpu_cores
    except Exception as e:
        print(f"Could not determine CPU cores: {e}")
        return 1  # Default to 1 if unable to determine




##############################################
#              Print Messages                #
##############################################

def print_startup_message():
    """
    Print the startup banner message for the Electrostatic Map Suite.
    
    """
    print(f"\n")
    print(f"="*60)
    print(f"                   Electrostatic Map Suite")
    print(f"                    By Stephen O. Ajagbe")
    print(f"="*60)


def print_office_quote():
    """
    Makes an API request to get a random quote and character from The Office TV show.
    Also cleans up any existing quote SVG files from the current directory.
    
    Raises:
        requests.exceptions.RequestException: If the API request fails
        
    Note:
        Uses the Office API at https://officeapi.akashrajpurohit.com
    """
    print("\nFetching inspirational quote...\n")
    url = f"{OFFICE_API}/quote/random"
    resp = requests.get(url)
    resp.raise_for_status()
    data = resp.json()
    #remove quote*.svg if it exists, * is a wildcard for any characters
    for file in os.listdir():
        if file.startswith("quote") and file.endswith(".svg"):
            try:
                os.remove(file)
            except OSError as e:
                print(f"Warning: could not remove {file}: {e}")
    print(f"\n  {data['quote']} \n                     - {data['character']}\n")




##############################################
#         Molecular Object Creation          #
##############################################

def create_molecule_object(
    atom_input,
    basis_set,
    method='dft',
    functional='b3lyp',
    original_charge=0,
    charge_change=0,
    gpu=True,
    spin_guesses=None #Note PySCF uses 2S = number of unpaired electrons not multiplicity (2S+1), so spin=0 is singlet, spin=1 is doublet, etc.
):
    """
    Create a PySCF molecule object with optimal spin configuration.
    
    This function creates a PySCF molecule object by testing different spin states
    and returns the one with the lowest energy. It supports both DFT and HF methods
    and can utilize GPU acceleration if available.
    
    Args:
        atom_input (str or list): Atomic coordinates (XYZ file path or coordinate list)
        basis_set (str): Basis set name (e.g., 'sto-3g', '6-31g*'), a list is provided in the method-info basis-sets file
        method (str, optional): Ab initio method ('dft' or 'hf'). Defaults to 'dft'.
        functional (str, optional): DFT functional name. Defaults to 'b3lyp',however, an extensive list is provided in the method-info functionals csv file with codes for easy access 
                                    e.g HYB_GGA_XC_WB97X_D3 can also be used with code 399.
        original_charge (int, optional): Base molecular charge. Defaults to 0.
        charge_change (int, optional): Charge modification. Defaults to 0. Useful for generating ions.
        gpu (bool, optional): Use GPU acceleration if available. Defaults to True.
        spin_guesses (list, optional): List of spin multiplicities to test. 
                                     Defaults to [0, 1, 2, 3, 4]. Uses 2S notation not multiplicity (2S+1).
                                     Important for open-shell systems.
    
    Returns:
        pyscf.scf object gpu4pyscf.scf object or None: The converged SCF object with lowest energy,
                                 or None if no spin state converged.
    
    Note:
        - Uses 2S notation (0=singlet, 1=doublet, etc.)
        - Automatically tries SOSCF if initial SCF doesn't converge
        - Prints convergence information for each spin state
    """
    charge = original_charge + charge_change
    spin_guesses = spin_guesses or [0, 1, 2, 3, 4]
    results = []  # store (spin, energy, mf)

    for spin in spin_guesses:
        try:
            mol = gto.Mole()
            mol.atom = atom_input
            mol.basis = basis_set
            mol.charge = charge
            mol.spin = spin  
            mol.build()

            # RKS for singlet, UKS for open shell
            if method.lower() == 'dft':
                mol.xc = functional
                mf = dft.UKS(mol) if spin > 0 else dft.RKS(mol)
            elif method.lower() == 'hf':
                mf = scf.UHF(mol) if spin > 0 else scf.RHF(mol)
            else:
                raise ValueError("Method must be 'dft' or 'hf'")

            # Move to GPU if available and requested
            if gpu:
                mf = mf.to_gpu()
            else:
                print("GPU not available or not requested - using CPU.")
            energy = mf.kernel()

            # Try SOSCF if not converged
            if not mf.converged:
                mf = mf.newton()
                energy = mf.kernel()
            
            if mf.converged:
                print(f"Spin {spin} (2S+1={spin+1}) converged: E = {energy:.6f} Ha")
                results.append((spin, energy, mf))
            else:
                print(f"Spin {spin} (2S+1={spin+1}) did NOT converge")

        except Exception as e:
            print(f"Spin {spin} failed: {e}")

    if results:
        # pick lowest energy among converged spins
        best_spin, best_energy, best_mf = min(results, key=lambda x: x[1])
        print(f"\nLowest energy: spin={best_spin} (2S+1={best_spin+1}), E={best_energy:.6f} Ha")
        return best_mf
    else:
        print("No spin converged for this species.")
        return None




##############################################
#        Molecular Object Manipulation       #
##############################################

def solvate_molecule(mf, solvent='water'):
    """
    Apply implicit solvation to a molecule using the Polarizable Continuum Model (PCM).
    
    This function adds solvation effects to an existing SCF object using the 
    C-PCM method with SMD solvent parameters.
    
    Args:
        mf (pyscf.scf object): The molecular SCF object to solvate
        solvent (str, optional): Solvent name from SMD database. Defaults to 'water'.
    
    Returns:
        pyscf.scf object: The solvated SCF object
        
    Note:
        - Uses C-PCM method with Lebedev order 29 for cavity construction
        - Automatically tries SOSCF if initial SCF doesn't converge
        - Solvent parameters are taken from the PySCF SMD database
    """
    solvent = solvent.lower()
    mf = mf.PCM()
    mf.with_solvent.eps = smd.solvent_db[solvent][5]
    mf.with_solvent.method = 'C-PCM'
    mf.with_solvent.lebedev_order = 29
    mf.kernel()
    if not mf.converged:
        print("SCF did not converge with solvent model. Trying SOSCF...")
        mf = mf.newton()
        mf.kernel()
        if not mf.converged:
            print("SOSCF also did not converge.")
        else:
            print("SOSCF converged.")
    return mf

def save_chkfile(mf, filename):
    """
    Save a PySCF calculation to a checkpoint file.
    
    This function creates a checkpoint file containing the molecular orbital
    information and other calculation data for later use.
    
    Args:
        mf (pyscf.scf object): The SCF object to save
        filename (str): Base filename for the checkpoint file (i.e. without .chk extension)
        
    Note:
        The function automatically adds the .chk extension to the filename.
    """
    mf.chkfile = f'{filename}.chk'
    mf.dump_chk(mf.__dict__)
    print(f"Saved checkpoint to {filename}.chk")

def find_homo_lumo_and_gap(mf):
    """
    Calculate HOMO, LUMO energies and the HOMO-LUMO gap from an SCF object.
    
    This function analyzes the molecular orbitals to identify the highest occupied
    molecular orbital (HOMO) and lowest unoccupied molecular orbital (LUMO).
    
    Args:
        mf (pyscf.scf object): Converged SCF object containing molecular orbitals
        
    Returns:
        tuple: (HOMO energy, LUMO energy, HOMO-LUMO gap) in eV

    Note:
        - HOMO is the highest energy orbital with non-zero occupation
        - LUMO is the lowest energy orbital with zero occupation
        - Gap is calculated as LUMO - HOMO
    """
    homo = -float("inf")
    lumo = float("inf")
    for energy, occ in zip(mf.mo_energy, mf.mo_occ):
        if occ > 0 and energy > homo:
            homo = energy 
        if occ == 0 and energy < lumo:
            lumo = energy 
    return homo, lumo, lumo - homo

def create_qmmm_molecule_object(mf, coord_mm, q_mm, chkfile = None):
    """
    Create a QM/MM (Quantum Mechanics/Molecular Mechanics) calculation object.
    
    This function integrates classical point charges (MM region) with a quantum
    mechanical calculation. It handles both CPU and GPU-based SCF objects.
    
    Args:
        mf (pyscf.scf or gpu4pyscf.scf object): Base SCF object for the QM region
        coord_mm (numpy.ndarray): MM coordinates with shape [N, 3]
        q_mm (numpy.ndarray): MM point charges with shape [N]
        chkfile (str, optional): Checkpoint file for initial guess. Defaults to None.
                                 Typically mf's chkfile, useful for quick convergence.

    Returns:
        pyscf.scf or gpu4pyscf.scf object: New SCF object with MM charges integrated
        
    Note:
        - For GPU calculations, MM integration is performed on CPU then transferred
        - Automatically tries SOSCF if initial SCF doesn't converge
        - The MM charges modify both the core Hamiltonian and nuclear repulsion energy
    """
    if not hasattr(mf, 'to_cpu'):
        mf_new = qmmm.mm_charge(mf, coord_mm, q_mm)
    else:
        mol = mf.mol        
        # Create a new SCF object of the same type and settings
        mf_new = type(mf)(mol)
        mf_new.__dict__.update(mf.__dict__)     
        # Move to CPU for MM charge integration
        temp_mf = mf_new.to_cpu()       
        temp_mf_mm = qmmm.mm_charge(temp_mf, coord_mm, q_mm)
        v_mm = temp_mf_mm.get_hcore() - temp_mf.get_hcore()
        e_nuc_mm = temp_mf_mm.energy_nuc() - temp_mf.energy_nuc()
        v_mm_gpu = cp.asarray(v_mm)
        orig_get_hcore = mf_new.get_hcore
        orig_energy_nuc = mf_new.energy_nuc
        def get_hcore_with_mm(*args):
            hcore = orig_get_hcore()
            return hcore + v_mm_gpu
        def energy_nuc_with_mm(*args):
            return orig_energy_nuc() + e_nuc_mm
        mf_new.get_hcore = get_hcore_with_mm
        mf_new.energy_nuc = energy_nuc_with_mm      
        mf_new.charge = mol.charge
        mf_new.spin = mol.spin
        mf_new.basis = mol.basis
        if hasattr(mf, 'xc'):
            mf_new.xc = mf.xc

    if chkfile:
        mf_new.chkfile = chkfile
        mf_new.init_guess = 'chkfile'
    mf_new.kernel()
    if not mf_new.converged:
        print("SCF did not converge with MM charges. Trying SOSCF...")
        mf_new = mf_new.newton()
        mf_new.kernel()
        if not mf_new.converged:
            print("SOSCF also did not converge.")
        else:
            print("SOSCF converged.")
    return mf_new


def create_td_molecule_object(mf, nstates=5, triplet=False):
    """
    Create a time-dependent (TD) calculation object for excited states.
    
    This function creates either a TDDFT or TDHF object from a converged ground
    state calculation to compute excited state properties and electronic transitions.
    
    Args:
        mf (pyscf.scf or gpu4pyscf.scf object): Converged ground state SCF object
        nstates (int, optional): Number of excited states to calculate. Defaults to 5.
        triplet (bool, optional): Calculate triplet states if True, singlet if False.
                                 Defaults to False.
    
    Returns:
        pyscf.tdscf object: Time-dependent calculation object with computed excited states
        
    Note:
        - Automatically detects DFT vs HF ground state from mf attributes and uses appropriate TD method
        - Handles both solvated and non-solvated molecules
        - Compatible with both CPU and GPU calculations
        
    Raises:
        ValueError: If the ground state object type is not supported
    """
    if hasattr(mf, 'to_cpu') and callable(mf.to_cpu):
        if hasattr(mf, 'TDDFT') and mf.TDDFT is not None:
            td = mf.TDDFT()
        elif hasattr(mf, 'TDHF') and mf.TDHF is not None:
            td = mf.TDHF()
        else:
            raise ValueError("Unsupported ground state object type")
    else:   
        # Handle solvated molecules
        if hasattr(mf, 'with_solvent'):
            # For solvated molecules, use tdscf directly
            if hasattr(mf, 'xc'):  # DFT case
                td = tdscf.TDDFT(mf)
            else:  # HF case
                td = tdscf.TDHF(mf)
        else:
            # For non-solvated molecules, use the original method
            if hasattr(mf, 'TDDFT') and mf.TDDFT is not None:
                td = mf.TDDFT()
            elif hasattr(mf, 'TDHF') and mf.TDHF is not None:
                td = mf.TDHF()
            else:
                raise ValueError("Unsupported ground state object type")

    td.singlet = not triplet
    td.nstates = nstates
    td.kernel()
    return td


##############################################
#          Molecular File Operations         #
##############################################

def get_vdw_surface_coordinates(xyz_file):
    """
    Generate van der Waals surface coordinates for a molecule.
    
    This function uses the external 'vsg' tool to generate points on the 
    van der Waals surface of a molecule from its XYZ coordinates.
    
    Args:
        xyz_file (str): Path to the XYZ file containing molecular coordinates
        
    Returns:
        numpy.ndarray: Array of surface coordinates with shape [N, 3]
        
    Raises:
        RuntimeError: If the vsg command fails
        FileNotFoundError: If the expected surface file is not created
        
    Note:
        - Requires the 'vsg' external tool to be installed and in PATH
        - Automatically cleans up temporary surface files after reading
        - Handles single-point surfaces by reshaping to proper dimensions
    """
    ret = subprocess.run(['vsg', xyz_file, '--txt'], capture_output=True, text=True)
    if ret.returncode != 0:
        raise RuntimeError(f"vsg failed: {ret.stderr}")
    base, _ = os.path.splitext(xyz_file)
    surface_file = f"{base}_vdw_surface.txt"
    if not os.path.isfile(surface_file):
        raise FileNotFoundError(f"Expected surface file not found: {surface_file}")
    coords = np.loadtxt(surface_file, dtype=float)
    if coords.ndim == 1 and coords.size == 3:
        coords = coords.reshape(1,3)
    try:
        os.remove(surface_file)
    except OSError as e:
        print(f"Warning: could not remove {surface_file}: {e}")
    return coords

def extract_xyz_name(xyz_filepath):
    """
    Extract a clean molecule name from an XYZ file path.
    
    This function takes a file path and returns the base filename without
    extension, with path separators replaced by underscores for safe filename usage.
    
    Args:
        xyz_filepath (str): Path to the XYZ file
        
    Returns:
        str: Clean molecule name suitable for use in output filenames
        
    Note:
        Replaces both forward slashes and backslashes with underscores
        to ensure cross-platform compatibility.
    """
    molecule_name = os.path.splitext(os.path.basename(xyz_filepath))[0]
    molecule_name = molecule_name.replace('/', '_').replace('\\', '_')
    return molecule_name

def optimize_molecule(xyz_filepath, 
    basis_set,
    method='dft',
    functional='b3lyp',
    original_charge=0,
    charge_change=0,
    gpu=True,
    spin_guesses=None):
    """
    Perform geometry optimization on a molecule and save the optimized structure.
    
    This function takes a molecular structure from an XYZ file, creates a quantum
    mechanical calculation object, optimizes the geometry, and writes the optimized
    coordinates to a new XYZ file.
    
    Args:
        xyz_filepath (str): Path to input XYZ file with initial geometry
        basis_set (str): Basis set name (e.g., 'sto-3g', '6-31g*'), a list is provided in the method-info basis-sets file
        method (str, optional): Ab initio method ('dft' or 'hf'). Defaults to 'dft'.
        functional (str, optional): DFT functional name. Defaults to 'b3lyp',however, an extensive list is provided in the method-info functionals csv file with codes for easy access 
                            e.g HYB_GGA_XC_WB97X_D3 can also be used with code 399.
        original_charge (int, optional): Base molecular charge. Defaults to 0.
        charge_change (int, optional): Charge modification. Defaults to 0. Useful for generating ions.
        gpu (bool, optional): Use GPU acceleration if available. Defaults to True.
        spin_guesses (list, optional): List of spin multiplicities to test. 
                             Defaults to [0, 1, 2, 3, 4]. Uses 2S notation not multiplicity (2S+1).
                             Important for open-shell systems.
    
    Returns:
        str: Filename of the output XYZ file containing optimized geometry
        
    Raises:
        ValueError: If molecule object creation fails
        
    Note:
        - Uses PySCF's geometric solver with convergence tolerance of 1e-7
        - Output filename format: "{molecule_name}_opt.xyz"
        - Coordinates are written in Angstrom units
    """
    
    molecule_name = extract_xyz_name(xyz_filepath)
    
    # Create molecule object using the parameters
    mf = create_molecule_object(
        atom_input=xyz_filepath,
        basis_set=basis_set,
        method=method,
        functional=functional,
        original_charge=original_charge,
        charge_change=charge_change,
        gpu=gpu,
        spin_guesses=spin_guesses
    )
    
    if mf is None:
        raise ValueError("Failed to create molecule object")
    
    # Optimize geometry
    mol_eq = optimize(mf, conv_tol=1e-7)
    coords = mol_eq.atom_coords(unit='Ang')
    atoms = [mol_eq.atom_symbol(i) for i in range(mol_eq.natm)]
    
    # Write to XYZ file
    output_filename = f"{molecule_name}_opt.xyz"
    with open(output_filename, 'w') as f:
        # Write number of atoms
        f.write(f"{len(atoms)}\n")
        # Write comment line
        f.write("Optimized geometry from PySCF\n")
        # Write atom coordinates
        for atom, coord in zip(atoms, coords):
            f.write(f"{atom:2s} {coord[0]:12.8f} {coord[1]:12.8f} {coord[2]:12.8f}\n")
    
    print(f"Optimized geometry written to {output_filename}")
    return output_filename


def smiles_to_xyz(smiles, filename=None):
    """
    Convert a SMILES string to an XYZ coordinate file.
    
    This function uses RDKit to generate 3D coordinates from a SMILES string,
    including hydrogen atoms and basic geometry optimization using MMFF.
    
    Args:
        smiles (str): SMILES representation of the molecule
        filename (str, optional): Output filename. If None, generates automatic name.
                                 Defaults to None.
    
    Returns:
        str: Path to the generated XYZ file
        
    Note:
        - Automatically adds hydrogen atoms to the molecule
        - Performs 3D embedding and MMFF geometry optimization
        - If no filename provided, uses format "mol_{hash}.xyz"
        - Hash is generated from SMILES string for reproducibility
    """
    mol = Chem.AddHs(Chem.MolFromSmiles(smiles))
    AllChem.EmbedMolecule(mol)
    AllChem.MMFFOptimizeMolecule(mol)
    
    if not filename:
        filename = f"mol_{abs(hash(smiles)) % 10000}.xyz"
    
    conf = mol.GetConformer()
    with open(filename, 'w') as f:
        f.write(f"{mol.GetNumAtoms()}\n{smiles}\n")
        for i, atom in enumerate(mol.GetAtoms()):
            pos = conf.GetAtomPosition(i)
            f.write(f"{atom.GetSymbol()} {pos.x:.6f} {pos.y:.6f} {pos.z:.6f}\n")
    return filename



def create_mol2_file(molecule_name, coordinates, property_list, property_name):
    """
    Create a MOL2 file with molecular property data mapped to coordinates.
    
    This function generates a MOL2 format file where each coordinate point
    is assigned a property value. The file can be used for visualization
    of molecular properties in molecular graphics software.
    
    Args:
        molecule_name (str): Base name for the molecule
        coordinates (array-like): Array of 3D coordinates with shape [N, 3]
        property_list (array-like): Property values for each coordinate point with shape [N]
        property_name (str): Name/type of the property being mapped
        
    Returns:
        None: Writes output directly to file
        
    Raises:
        ValueError: If coordinates and property_list have different lengths
        
    Note:
        - Output filename format: "{molecule_name}_{property_name}.mol2"
        - Each point is represented as a hydrogen atom for visualization
        - Property values are stored in the charge field of the MOL2 format
        - Uses TRIPOS MOL2 format specification
    """
    if len(coordinates) != len(property_list):
        raise ValueError("coordinates and property_list must have same length")

    filename = f"{molecule_name}_{property_name}.mol2"
    num_points = len(coordinates)
    
    with open(filename, 'w') as f:
        f.write("@<TRIPOS>MOLECULE\n")
        f.write(f"{filename}\n")
        f.write(f"    {num_points} 0 0 0\n")   
        f.write("SMALL\n")
        f.write("GASTEIGER\n")
        f.write("@<TRIPOS>ATOM\n")
        
        for i, (coord, prop_val) in enumerate(zip(coordinates, property_list), start=1):
            x, y, z = coord
            atom_name = "H"
            atom_type = "H1"
            subst_id = 1
            subst_name = property_name.upper()
            f.write(f"{i:>4} {atom_name:<4} {x:>9.4f} {y:>9.4f} {z:>9.4f} {atom_type:<4} {subst_id} {subst_name} {prop_val:>10.6f}\n")





import numpy as np
import cupy as cp
from pyscf import gto, scf, dft, tdscf, qmmm, lib
from pyscf.solvent import smd
from pyscf.geomopt.geometric_solver import optimize


try:
    device_count = cp.cuda.runtime.getDeviceCount()
    GPU_AVAILABLE = device_count > 0
    
    if GPU_AVAILABLE:
        print(f"\n Found {device_count} GPU(s):")
        for i in range(device_count):
            cp.cuda.Device(i).use()
            props = cp.cuda.runtime.getDeviceProperties(i)
            name = props['name'].decode()
            print(f"  GPU {i}: {name}")        
        # Test the first GPU
        cp.cuda.Device(0).use()
        test_array = cp.array([1, 2, 3])
        print(f"\n Using GPU 0 for computations. \n")
    else:
        print("\n No GPUs found.\n \n Switching to CPU mode.\n")
        
except (ImportError, cp.cuda.runtime.CUDARuntimeError) as e:
    GPU_AVAILABLE = False
    print(f"GPU not available: {e}")

# if GPU_AVAILABLE:



# else:



def create_qmmm_molecule_object(mf, coord_mm, q_mm, chkfile = None):
    """
    mf: SCF object (already created, possibly converged without MM)
    coord_mm: array of MM coordinates (shape [N,3])
    q_mm: array of MM charges (shape [N])

    """
    # mf_new = qmmm.mm_charge(mf, coord_mm, q_mm)
    if not hasattr(mf, 'to_cpu'):
        mf_new = qmmm.mm_charge(mf, coord_mm, q_mm)
    else:
        mol = mf.mol        
        # Create a new SCF object of the same type and settings
        mf_new = type(mf)(mol)
        mf_new.__dict__.update(mf.__dict__)     
        # Move to CPU for MM charge integration
        temp_mf = mf_new.to_cpu()       
        temp_mf_mm = qmmm.mm_charge(temp_mf, coord_mm, q_mm)
        v_mm = temp_mf_mm.get_hcore() - temp_mf.get_hcore()
        e_nuc_mm = temp_mf_mm.energy_nuc() - temp_mf.energy_nuc()
        v_mm_gpu = cp.asarray(v_mm)
        orig_get_hcore = mf_new.get_hcore
        orig_energy_nuc = mf_new.energy_nuc
        def get_hcore_with_mm(*args):
            hcore = orig_get_hcore()
            return hcore + v_mm_gpu
        def energy_nuc_with_mm(*args):
            return orig_energy_nuc() + e_nuc_mm
        mf_new.get_hcore = get_hcore_with_mm
        mf_new.energy_nuc = energy_nuc_with_mm      
        mf_new.charge = mol.charge
        mf_new.spin = mol.spin
        mf_new.basis = mol.basis
        if hasattr(mf, 'xc'):
            mf_new.xc = mf.xc

    if chkfile:
        mf_new.chkfile = chkfile
        mf_new.init_guess = 'chkfile'
    mf_new.kernel()
    if not mf_new.converged:
        print("SCF did not converge with MM charges. Trying SOSCF...")
        mf_new = mf_new.newton()
        mf_new.kernel()
        if not mf_new.converged:
            print("SOSCF also did not converge.")
        else:
            print("SOSCF converged.")
    return mf_new




def create_charged_molecule_object(
    atom_input,
    basis_set,
    method='dft',
    functional='b3lyp',
    original_charge=0,
    charge_change=0,
    gpu=False,
    spin_guesses=None,
):
    charge = original_charge + charge_change
    spin_guesses = spin_guesses or [0, 1, 2, 3, 4]

    results = []  # store (spin, energy, mf)

    for spin in spin_guesses:
        try:
            mol = gto.Mole()
            mol.atom = atom_input
            mol.basis = basis_set
            mol.charge = charge
            mol.spin = spin   # number of unpaired electrons (2S)
            mol.build()

            # RKS for singlet, UKS for open shell
            if method.lower() == 'dft':
                mol.xc = functional
                mf = dft.UKS(mol) if spin > 0 else dft.RKS(mol)
            elif method.lower() == 'hf':
                mf = scf.UHF(mol) if spin > 0 else scf.RHF(mol)
            else:
                raise ValueError("Method must be 'dft' or 'hf'")
            if gpu:
                mf = mf.to_gpu()
            energy = mf.kernel()

            if not mf.converged:
                mf = mf.newton()
                energy = mf.kernel()
            
            if mf.converged:
                print(f"Spin {spin} (2S+1={spin+1}) converged: E = {energy:.6f} Ha")
                results.append((spin, energy, mf))
            else:
                print(f"Spin {spin} (2S+1={spin+1}) did NOT converge")

        except Exception as e:
            print(f"Spin {spin} failed: {e}")

    if results:
        # pick lowest energy among converged spins
        best_spin, best_energy, best_mf = min(results, key=lambda x: x[1])
        print(f"\nLowest energy: spin={best_spin} (2S+1={best_spin+1}), E={best_energy:.6f} Ha")
        return best_mf, best_spin, best_energy
    else:
        print("No spin converged for this species.")
        return None, None, None


def find_homo_lumo_and_gap(mf):
    homo = -float("inf")
    lumo = float("inf")
    for energy, occ in zip(mf.mo_energy, mf.mo_occ):
        if occ > 0 and energy > homo:
            homo = energy 
        if occ == 0 and energy < lumo:
            lumo = energy 
    return homo, lumo, lumo - homo

def solvate_molecule(mf, solvent='water'):
    solvent = solvent.lower()
    mf = mf.PCM()
    mf.with_solvent.eps = smd.solvent_db[solvent][5]
    mf.with_solvent.method = 'C-PCM'
    mf.with_solvent.lebedev_order = 29
    return mf

def optimize_and_get_equilibrium(mf):
    """
    Optimize the geometry of the molecule and return the equilibrium geometry.
    """
    mf = mf.to_cpu()
    mol_eq = optimize(mf,conv_tol_grad=1e-7,conv_tol=1e-10)
    coords = mol_eq.atom_coords(unit='Ang')
    atoms = [mol_eq.atom_symbol(i) for i in range(mol_eq.natm)]
    atom_list = [(atom, coord) for atom, coord in zip(atoms, coords)]
    return atom_list

molecule = "water.xyz"
coord_mm = np.array([[0.0, 0.0, 1.5]])
q_mm = np.array([1.0])

mf_gpu, _, _ = create_charged_molecule_object(
    atom_input=molecule,
    basis_set="6-31g*",
    method='hf',
    functional='b3lyp',
    original_charge=0,
    charge_change=0,
    gpu=True,
    spin_guesses=[0]
)


mf_gpu.chkfile = 'anion.chk'
mf_gpu.dump_chk(mf_gpu.__dict__)


mf_gpu_qmmm = create_qmmm_molecule_object(mf_gpu, coord_mm, q_mm, chkfile = 'anion.chk')



solvated_vac = solvate_molecule(mf_gpu, solvent='water')
solvated_wsc = solvate_molecule(mf_gpu_qmmm, solvent='water')


solvated_vac.kernel()
solvated_wsc.kernel()

print(mf_gpu_qmmm.e_tot, mf_gpu.e_tot)
print(solvated_wsc.e_tot,solvated_vac.e_tot)


homo, lumo, gap = find_homo_lumo_and_gap(mf_gpu)
homo_wsc, lumo_wsc, gap_wsc = find_homo_lumo_and_gap(mf_gpu_qmmm)

homo_sol_vac, lumo_sol_vac, gap_sol_vac = find_homo_lumo_and_gap(solvated_vac)
homo_sol_wsc, lumo_sol_wsc, gap_sol_wsc = find_homo_lumo_and_gap(solvated_wsc)



print(f"VAC: {homo}, {lumo}, {gap}")
print(f"WSC: {homo_wsc}, {lumo_wsc}, {gap_wsc}")
print(f"Solvated VAC: {homo_sol_vac}, {lumo_sol_vac}, {gap_sol_vac}")
print(f"Solvated WSC: {homo_sol_wsc}, {lumo_sol_wsc}, {gap_sol_wsc}")

# #TDSCF Properties




nstates = 5



td_vac = mf_gpu.TDDFT()
td_vac.nstates = nstates
td_vac.singlet = False
td_vac.kernel()


td_sol_vac = solvated_vac.TDDFT()
td_sol_vac.nstates = nstates
td_sol_vac.singlet = True
td_sol_vac.kernel()

td_wsc = mf_gpu_qmmm.TDDFT()
td_wsc.nstates = nstates
td_wsc.singlet = True   
td_wsc.kernel()

td_sol_wsc = solvated_wsc.TDDFT()
td_sol_wsc.nstates = nstates
td_sol_wsc.singlet = True
td_sol_wsc.kernel()


print("Excitation energies (Ha):", td_vac.e)
print("Oscillator strengths:", td_vac.oscillator_strength())

print("Excitation energies (Ha):", td_sol_vac.e)
print("Oscillator strengths:", td_sol_vac.oscillator_strength())

print("Excitation energies (Ha):", td_wsc.e)
print("Oscillator strengths:", td_wsc.oscillator_strength())

print("Excitation energies (Ha):", td_sol_wsc.e)
print("Oscillator strengths:", td_sol_wsc.oscillator_strength())


# srun -p qGPU48 -A CHEM9C4 --nodes=1 --ntasks-per-node=1 --cpus-per-task=1 --gres=gpu:1 --time=01:00:00 --mem=8G --pty bash





# def gpu_worker(gpu_id, calculation_func, result_dict, key, *args):
#     """Worker function for GPU calculations"""
#     with cp.cuda.Device(gpu_id):  # Switch to specific GPU
#         result_dict[key] = calculation_func(*args)  # Run calc and store result


# def run_scf_calculation(mol_data):
#     """Example SCF calculation function"""
#     mol = gto.M(**mol_data)
#     mf = scf.RHF(mol).to_gpu()
#     return mf.kernel()

# if No_of_GPUs:
#     # Run two different molecules simultaneously




    
#     mol1_data = {'atom': 'H 0 0 0; H 0 0 0.74', 'basis': 'cc-pVDZ'}
#     mol2_data = {'atom': 'He 0 0 0', 'basis': 'cc-pVDZ'}
    
#     results = {}
#     t1 = threading.Thread(target=gpu_worker, 
#                          args=(0, run_scf_calculation, results, 'H2', mol1_data))
#     t2 = threading.Thread(target=gpu_worker, 
#                          args=(1, run_scf_calculation, results, 'He', mol2_data))
    
#     t1.start()
#     t2.start()
#     t1.join()
#     t2.join()
    
#     print(f"H2 energy: {results['H2']}")
#     print(f"He energy: {results['He']}")



# def get_redox_potential(molecule, gfec_basis_set, gfec_functional, sscf_basis_set, sscf_functional, charge, spin, method, solvent):
# #Neutral
#     rdx_n = create_molecule_object(molecule.mol.atom, basis_set=gfec_basis_set, method=method,functional=gfec_functional, charge=charge,spin=spin)
#     rdx_n_opt = optimize_and_get_equilibrium(rdx_n)
#     rdx_n = create_molecule_object(rdx_n_opt,basis_set=gfec_basis_set, method=method,functional=gfec_functional, charge=charge,spin=spin)
#     _, _, G_corr_rdx_n = get_molecule_gfec(rdx_n)

#     # #Anion
#     rdx_a, _, _ = create_charged_molecule_object(molecule.mol.atom, basis_set=gfec_basis_set, method=method, functional=gfec_functional, original_charge=charge, original_spin=spin, charge_change=-1)
#     rdx_a_opt = optimize_and_get_equilibrium(rdx_a)
#     rdx_a,_,_= create_charged_molecule_object(rdx_a_opt, basis_set=gfec_basis_set, method=method, functional=gfec_functional, original_charge=charge, original_spin=spin, charge_change=-1)
#     _,_, G_corr_rdx_a = get_molecule_gfec(rdx_a)

#     # #hbs = higher basis set solvated SCF
#     rdx_n_hbs = create_molecule_object(rdx_n.mol.atom, basis_set=basis_set, functional=functional, method=method, charge=charge, spin=spin)
#     solvated_n = solvate_molecule(rdx_n_hbs, solvent=solvent)
    
#     rdx_a_hbs,_,_ = create_charged_molecule_object(rdx_a.mol.atom, basis_set=basis_set, functional=functional, method=method, original_charge=charge, original_spin=spin, charge_change=-1)
#     solvated_a = solvate_molecule(rdx_a_hbs, solvent=solvent)

#     solvated_n.kernel()
#     solvated_a.kernel()
    
#     #GFE Adjustment
#     gfe_n = solvated_n.e_tot + G_corr_rdx_n
#     gfe_a = solvated_a.e_tot + G_corr_rdx_a

#     #Redox Gibbs Free Energy 
#     dG_red  = -1*((gfe_a - gfe_n) * HARTREE_TO_EV)
#     rdx_pot = (dG_red) - SHE_REFERENCE_POTENTIAL

#     print(f"G_corr_rdx_a: {G_corr_rdx_a}, G_corr_rdx_n: {G_corr_rdx_n}")
#     print(f"gfe_a: {gfe_a}, gfe_n: {gfe_n}, dG_red: {dG_red}, rdx_pot: {rdx_pot}")

#     return rdx_pot


# get_redox_potential(molecule, gfec_basis_set, gfec_functional, sscf_basis_set=basis_set, sscf_functional=functional, charge=charge, spin=spin, method=method, solvent=solvent)




import core,csv,os,argparse,ast,sys
import numpy as np


##############################################
#              Dependency Setup              #
##############################################

HARTREE_TO_KCAL = 627.509
HARTREE_TO_EV = 27.2114


PROPERTY_CONFIG = {
    'gse': {'deps': [], 'calc': [], 'unit': 1},
    'homo': {'deps': [], 'calc': [], 'unit': 1},
    'lumo': {'deps': [], 'calc': [], 'unit': 1},
    'gap': {'deps': ['homo', 'lumo'], 'calc': [], 'unit': 1},
    'dm': {'deps': [], 'calc': [], 'unit': 1},
    'ie': {'deps': [], 'calc': ['cation'], 'unit': HARTREE_TO_KCAL},
    'ea': {'deps': [], 'calc': ['anion'], 'unit': HARTREE_TO_KCAL},
    'cp': {'deps': ['ie', 'ea'], 'calc': [], 'unit': HARTREE_TO_KCAL},
    'eng': {'deps': ['cp'], 'calc': [], 'unit': HARTREE_TO_EV},
    'hard': {'deps': ['ie', 'ea'], 'calc': [], 'unit': HARTREE_TO_EV},
    'efl': {'deps': ['cp', 'hard'], 'calc': [], 'unit': HARTREE_TO_EV},
    'nfl': {'deps': ['efl'], 'calc': [], 'unit': HARTREE_TO_EV},
    'exe': {'deps': [], 'calc': ['td'], 'unit': 1},
    'osc': {'deps': [], 'calc': ['td'], 'unit': 1}
}


#####################################################
# Prepare Calculations' Dependencies on User Inputs #
#####################################################

def setup_calculation(requested_props):
    """
    Setup and resolve property dependencies and required calculations.
    
    This function takes a list of requested molecular properties and determines
    all the dependencies and quantum mechanical calculations needed to compute them.
    It handles the dependency tree resolution and maps properties to required
    calculations (neutral, cation, anion, TD).
    
    Args:
        requested_props (list): List of property names to calculate.
                               Use 'all' to calculate all available properties.
                               
    Returns:
        tuple: (props_needed, calcs_needed) where:
            - props_needed (list): All properties needed including dependencies
            - calcs_needed (dict): Dictionary mapping calculation types to boolean
                                  (e.g., {'neutral': True, 'cation': False, ...})
    
    Note:
        Available properties: 'gse', 'homo', 'lumo', 'gap', 'dm', 'ie', 'ea', 
        'cp', 'eng', 'hard', 'efl', 'nfl', 'exe', 'osc' and the all encompassing 'all'.
        
        Dependencies are automatically resolved (e.g., 'gap' requires 'homo' and 'lumo')
    """
    if 'all' in requested_props:
        requested_props = list(PROPERTY_CONFIG.keys())
    
    # Resolve dependencies
    props_needed = set()
    def add_deps(prop):
        if prop in props_needed: return
        props_needed.add(prop)
        for dep in PROPERTY_CONFIG[prop]['deps']:
            add_deps(dep)
    
    for prop in requested_props:
        add_deps(prop)
    
    # Determine required calculations
    calcs_needed = {'neutral': True}
    for prop in props_needed:
        for calc in PROPERTY_CONFIG[prop]['calc']:
            calcs_needed[calc] = True
    
    return list(props_needed), calcs_needed


###############################################
# Calculate Properties from Molecular Objects #
###############################################

def calculate_all_properties(mf, anion_mf=None, cation_mf=None, td_obj=None, triplet=False, props_to_calc=None):
    """
    Calculate a comprehensive set of molecular properties from quantum calculations.
    
    This function computes various molecular properties including energetic,
    electronic, and excited state properties from converged SCF objects.
    
    Args:
        mf (pyscf.scf or gpu4pyscf.scf object): Converged neutral molecule SCF object
        anion_mf (pyscf.scf or gpu4pyscf.scf object, optional): Converged anion SCF object for EA calculations
        cation_mf (pyscf.scf or gpu4pyscf.scf object, optional): Converged cation SCF object for IE calculations  
        td_obj (pyscf.tdscf or gpu4pyscf.scf object, optional): TD object for excited state properties
        triplet (bool, optional): Whether to calculate triplet excited states. Defaults to False.
        props_to_calc (list, optional): List of properties to calculate
        
    Returns:
        dict: Dictionary containing calculated properties with units:
            - 'gse': Ground state energy (kcal/mol)
            - 'homo'/'lumo'/'gap': Orbital energies (eV)
            - 'dm': Dipole moment magnitude (Debye)
            - 'ie'/'ea': Ionization/electron affinity (kcal/mol)
            - 'cp': Chemical potential (kcal/mol)
            - 'eng': Electronegativity (eV)
            - 'hard': Chemical hardness (eV)
            - 'efl'/'nfl': Electrophilicity/nucleophilicity (eV)
            - 's1_exe'/'t1_exe': Excitation energies (eV)
            - 's1_osc'/'t1_osc': Oscillator strengths (dimensionless)
            
    Note:
        - Energies are converted from Hartree using conversion constants
        - Excited state properties are labeled with state prefix (s/t) and number
        - Missing SCF objects will skip dependent property calculations
    """
    results = {}
    
    # Basic properties
    if 'gse' in props_to_calc:
        results['gse'] = mf.e_tot * HARTREE_TO_KCAL
    
    if any(p in props_to_calc for p in ['homo', 'lumo', 'gap']):
        homo, lumo, gap = [x * HARTREE_TO_EV for x in core.find_homo_lumo_and_gap(mf)]
        results.update({p: v for p, v in zip(['homo', 'lumo', 'gap'], [homo, lumo, gap]) if p in props_to_calc})
    
    if 'dm' in props_to_calc:
        results['dm'] = np.linalg.norm(mf.dip_moment())
    
    # Charged state properties
    if 'ie' in props_to_calc and cation_mf:
        results['ie'] = cation_mf.e_tot - mf.e_tot
    if 'ea' in props_to_calc and anion_mf:
        results['ea'] = mf.e_tot - anion_mf.e_tot
    
    # Derived properties
    if 'cp' in props_to_calc and all(k in results for k in ['ie', 'ea']):
        results['cp'] = -(results['ie'] + results['ea']) / 2
    if 'eng' in props_to_calc and 'cp' in results:
        results['eng'] = -results['cp']
    if 'hard' in props_to_calc and all(k in results for k in ['ie', 'ea']):
        results['hard'] = (results['ie'] - results['ea']) / 2
    if 'efl' in props_to_calc and all(k in results for k in ['cp', 'hard']):
        results['efl'] = results['cp']**2 / (2 * results['hard']) if results['hard'] != 0 else 0
    if 'nfl' in props_to_calc and 'efl' in results:
        results['nfl'] = 1/results['efl'] if results['efl'] != 0 else 0
    
    # Excited state properties - only if TD object exists and exe/osc properties are requested
    if td_obj and any(p in props_to_calc for p in ['exe', 'osc']):
        state_prefix = 't' if triplet else 's'
        
        if 'exe' in props_to_calc:
            excitation_energies = td_obj.e * HARTREE_TO_EV
            for i, energy in enumerate(excitation_energies, 1):
                results[f'{state_prefix}{i}_exe'] = energy
        
        if 'osc' in props_to_calc:
            oscillator_strengths = td_obj.oscillator_strength()
            for i, osc in enumerate(oscillator_strengths, 1):
                results[f'{state_prefix}{i}_osc'] = osc
    
    return results

def calculate_surface_effect_at_point(base_molecules, coord, surface_charge, solvent, state_of_interest, triplet, properties_to_calculate, required_calculations):
    """
    Calculate the effect of a surface charge at a single coordinate point.
    
    This function computes how a point charge at a specific location affects
    various molecular properties by comparing calculations with and without
    the external charge.
    
    Args:
        base_molecules (list): [neutral_mf, anion_mf, cation_mf, td_obj] base objects
        coord (array-like): 3D coordinates [x, y, z] of the surface charge
        surface_charge (float): Magnitude of the point charge
        solvent (str or None): Solvent for implicit solvation
        state_of_interest (int): Number of excited states for TD calculations
        triplet (bool): Whether to calculate triplet excited states
        properties_to_calculate (list): List of molecular properties to compute
        required_calculations (dict): Dictionary specifying needed calculations
        
    Returns:
        dict: Dictionary of property effects with keys like '{property}_effect'
              containing the difference (with_charge - without_charge)
              
    Note:
        - Creates QM/MM calculations with the single point charge
        - Applies solvation if specified
        - Calculates the difference in properties due to the external charge
        - Effects show how the external charge modifies each property
    """
    molecule_alone, anion_alone, cation_alone, td_alone = base_molecules
    
    # Create single-point charge array
    single_coord = np.array([coord])
    q_mm = np.array([surface_charge])
    
    # Create QM/MM objects for this point
    molecule_wsc, anion_wsc, cation_wsc, td_wsc = create_wsc_objects(
        [molecule_alone, anion_alone, cation_alone, td_alone], 
        single_coord, q_mm, state_of_interest, triplet, required_calculations
    )
    
    # Apply solvation if needed
    if solvent:
        all_molecules = [molecule_alone, anion_alone, cation_alone, td_alone, 
                       molecule_wsc, anion_wsc, cation_wsc, td_wsc]
        all_molecules = apply_solvation(all_molecules, solvent, state_of_interest, triplet, required_calculations)
        molecule_alone, anion_alone, cation_alone, td_alone, \
        molecule_wsc, anion_wsc, cation_wsc, td_wsc = all_molecules
    
    # Calculate properties
    results = calculate_all_properties(molecule_alone, anion_mf=anion_alone, 
                                     cation_mf=cation_alone, td_obj=td_alone, 
                                     triplet=triplet, props_to_calc=properties_to_calculate)
    wsc_results = calculate_all_properties(molecule_wsc, anion_mf=anion_wsc, 
                                         cation_mf=cation_wsc, td_obj=td_wsc, 
                                         triplet=triplet, props_to_calc=properties_to_calculate)
    
    # Calculate differences
    effects = {}
    for prop in results:
        if prop in wsc_results:
            effects[f'{prop}_effect'] = wsc_results[prop] - results[prop]
    
    return effects


##########################################################
# Create Molecular Objects for Any Requested Calculation #
##########################################################
def create_alone_molecule_objects(charge_change, spin_guesses):
    """
    Create individual molecule objects with specified charge modifications.
    
    This is a wrapper function that creates PySCF molecule objects using
    the global parameters and applies charge modifications for ionic states.
    
    Args:
        charge_change (int): Charge modification from neutral state (+1 for cation, -1 for anion)
        spin_guesses (list or None): List of spin multiplicities to test, e.g [0,1] for singlet and doublet. If None, default of [0, 1, 2, 3, 4] are used.
        
    Returns:
        pyscf.scf object or None: Converged SCF object for the specified charge state
        
    Note:
        Uses global variables: input_data, basis_set, method, functional, charge, No_of_GPUs
    """
    return core.create_molecule_object(
        atom_input=input_data,
        basis_set=basis_set,
        method=method,
        functional=functional,
        original_charge=charge,
        charge_change=charge_change,
        gpu=True if No_of_GPUs > 0 else False,
        spin_guesses=spin_guesses
    )

def create_molecule_objects(required_calculations, state_of_interest, triplet):
    """
    Create and save all required molecule objects for property calculations.
    
    This function creates SCF objects for neutral, anionic, and cationic states
    as needed, along with time-dependent objects for excited state calculations.
    All objects are saved as checkpoint files for later use.
    
    Args:
        required_calculations (dict): Dictionary specifying which calculations are needed
                                    (e.g., {'neutral': True, 'anion': False, ...})
        state_of_interest (int): Number of excited states to calculate for TD
        triplet (bool): Whether to calculate triplet excited states
        
    Returns:
        list: [neutral_mf, anion_mf, cation_mf, td_obj] where elements are
              SCF/TD objects or None if not calculated
              
    Note:
        - Saves checkpoint files as 'molecule_alone.chk', 'anion_alone.chk', 'cation_alone.chk'
        - TD objects are only created if explicitly required
        - Charge states: neutral (0), anion (-1), cation (+1)
    """
    molecules = {}
    
    calc_configs = [('neutral', 0, None), ('anion', -1, None), ('cation', +1, None)]
    for name, charge_change, spin_guesses in calc_configs:
        if required_calculations.get(name, False):
            molecules[name] = create_alone_molecule_objects(charge_change, spin_guesses)
    
    # Only create TD if explicitly needed
    if required_calculations.get('td', False):
        molecules['td'] = core.create_td_molecule_object(molecules['neutral'], nstates=state_of_interest, triplet=triplet)
    
    chkfile_map = {'neutral': 'molecule_alone', 'anion': 'anion_alone', 'cation': 'cation_alone'}
    for key, filename in chkfile_map.items():
        if molecules.get(key):
            core.save_chkfile(molecules[key], filename)
    
    return [molecules.get(k) for k in ['neutral', 'anion', 'cation', 'td']]

def create_wsc_objects(molecules, coord, q_mm, state_of_interest, triplet, required_calculations):
    """
    Create QM/MM molecule objects with external point charges.
    
    This function creates QM/MM calculations by adding external point charges
    to the base molecule objects. 
    
    Args:
        molecules (list): [neutral_mf, anion_mf, cation_mf, td_obj] base objects
        coord (numpy.ndarray): Coordinates of external charges with shape [N, 3]
        q_mm (numpy.ndarray): Values of external point charges with shape [N]
        state_of_interest (int): Number of excited states for TD calculations
        triplet (bool): Whether to calculate triplet excited states
        required_calculations (dict): Dictionary specifying needed calculations
        
    Returns:
        list: [molecule_wsc, anion_wsc, cation_wsc, td_wsc] QM/MM objects
              where elements are SCF/TD objects or None if not calculated
              
    Note:
        - 'wsc' suffix indicates "with surface charge"
        - Uses checkpoint files from base calculations as initial guesses
        - TD objects are only created if explicitly required
    """
    molecule_alone, anion_alone, cation_alone, td_alone = molecules
    
    qmmm_configs = [
        ('molecule_wsc', molecule_alone, 'molecule_alone.chk'),
        ('anion_wsc', anion_alone, 'anion_alone.chk'),
        ('cation_wsc', cation_alone, 'cation_alone.chk')
    ]
    
    qmmm_objects = {}
    for name, mol, chkfile in qmmm_configs:
        if mol is not None:
            qmmm_objects[name] = core.create_qmmm_molecule_object(mol, coord, q_mm, chkfile)
    
    # Only create TD if explicitly needed
    if qmmm_objects.get('molecule_wsc') and required_calculations.get('td', False):
        qmmm_objects['td_wsc'] = core.create_td_molecule_object(
            qmmm_objects['molecule_wsc'], nstates=state_of_interest, triplet=triplet
        )
    
    return [qmmm_objects.get(k) for k in ['molecule_wsc', 'anion_wsc', 'cation_wsc', 'td_wsc']]

def apply_solvation(molecules, solvent, state_of_interest, triplet, required_calculations):
    """
    Apply implicit solvation to all molecule objects.
    
    This function applies solvation effects using the Polarizable Continuum Model
    to all provided molecule objects (both base and QM/MM calculations).
    
    Args:
        molecules (list): List of 8 molecule objects:
                         [molecule_alone, anion_alone, cation_alone, td_alone,
                          molecule_wsc, anion_wsc, cation_wsc, td_wsc]
        solvent (str or None): Solvent name from SMD database, or None for gas phase
        state_of_interest (int): Number of excited states for TD calculations
        triplet (bool): Whether to calculate triplet excited states  
        required_calculations (dict): Dictionary specifying needed calculations
        
    Returns:
        list: Solvated molecule objects in the same order as input,
              with new TD objects created from solvated ground states
              
    Note:
        - Returns original objects unchanged if solvent is None
        - Creates new TD objects from solvated ground states if needed
        - Maintains the same ordering and None values as input
    """
    if not solvent:
        return molecules
    
    molecule_alone, anion_alone, cation_alone, td_alone, molecule_wsc, anion_wsc, cation_wsc, td_wsc = molecules
    
    solvated = [core.solvate_molecule(mol, solvent) if mol else None 
                for mol in [molecule_alone, anion_alone, cation_alone, molecule_wsc, anion_wsc, cation_wsc]]
    
    # Only create TD objects if needed
    td_alone_new = None
    td_wsc_new = None
    
    if required_calculations.get('td', False):
        if solvated[0]:
            td_alone_new = core.create_td_molecule_object(solvated[0], nstates=state_of_interest, triplet=triplet)
        if solvated[3]:
            td_wsc_new = core.create_td_molecule_object(solvated[3], nstates=state_of_interest, triplet=triplet)
    
    return solvated[:3] + [td_alone_new] + solvated[3:] + [td_wsc_new]



##########################################################
#        Handling Input and Output Data And Files       #
##########################################################

def get_tuning_parameters(filepath='tuning.in'):
    """
    Search for tuning.in file and return parameters.
    
    Args:
        filepath (str): Path to tuning file, defaults to 'tuning.in'
        
    Returns:
        dict: Dictionary of tuning parameters
    """
    if not os.path.exists(filepath):
        return {}
    
    params = {}
    try:
        with open(filepath, 'r') as f:
            for line in f:
                line = line.strip()
                if line and not line.startswith('#') and '=' in line:
                    key, value = line.split('=', 1)
                    key = key.strip()
                    value = value.strip()
                    
                    # Parse value
                    try:
                        params[key] = ast.literal_eval(value)
                    except (ValueError, SyntaxError):
                        params[key] = value
    except Exception as e:
        print(f"Error parsing tuning.in file: {e}")
        return {}
    
    return params


def prepare_input_data(input_type, input_data):
    """
    Prepare molecular coordinates from different input formats.
    
    This function handles different input types and converts them to XYZ format where necessary.
    For XYZ file input, it simply returns the file path. For SMILES input, it generates 3D coordinates and saves
    them in an XYZ file. For SMILES input, it performs automatic geometry optimization.
    Dependent on global variables for optimization parameters and core module's optimize_molecule and smiles_to_xyz functions.
    Args:
        input_type (str): Type of input - 'xyz' for XYZ file path or 'smiles' for SMILES string
        input_data (str): Either path to XYZ file or SMILES string
        
    Returns:
        str: Path to XYZ file with molecular coordinates
        
    Raises:
        ValueError: If input_type is not 'xyz' or 'smiles'
        
    Note:
        - For SMILES input: converts to 3D coordinates and performs geometry optimization
        - For XYZ input: returns the file path unchanged
        - Uses global variables for optimization: basis_set, method, functional, charge, No_of_GPUs
    """
    if input_type.lower() == 'xyz':
        return input_data
    elif input_type.lower() == 'smiles':
        raw_xyz = core.smiles_to_xyz(input_data)
        optimized_xyz = core.optimize_molecule(raw_xyz,
        basis_set=basis_set,
        method=method,
        functional=functional,
        original_charge=charge,
        charge_change=0,
        gpu=True if No_of_GPUs > 0 else False,
        spin_guesses=None)
        return optimized_xyz
    else:
        raise ValueError("input_type must be 'file' or 'smiles'")

def create_output_files(surface_coords, all_effects, molecule_name, properties_to_calculate):
    """
    Create MOL2 files and CSV summary for surface effects analysis.
    
    This function generates visualization files for molecular property maps
    and a comprehensive CSV summary of all calculated surface effects.
    
    Args:
        surface_coords (numpy.ndarray): Array of surface coordinates with shape [N, 3]
        all_effects (list): List of effect dictionaries for each surface point
        molecule_name (str): Base name for output files
        properties_to_calculate (list): List of calculated molecular properties
        
    Returns:
        None: Creates files directly on disk
        
    Output Files:
        - MOL2 files: "{molecule_name}_{property}.mol2" for each property
          (suitable for visualization in molecular graphics software)
        - CSV file: "{molecule_name}_tuning_summary.csv" with all effects data
          
    Note:
        - MOL2 files map property effects to surface coordinates for visualization
        - CSV contains point index, coordinates, and all property effects
        - Files are saved in the current working directory
        - Property effects represent the change due to external charges
    """
    
    # Collect all effects by property
    property_effects = {}
    for prop in properties_to_calculate:
        prop_effect_key = f'{prop}_effect'
        property_effects[prop] = [effect.get(prop_effect_key, 0.0) for effect in all_effects]
    
    # Create MOL2 files for each property
    for prop, effect_values in property_effects.items():
        core.create_mol2_file(molecule_name, surface_coords, effect_values, prop)
        # print(f"Created {molecule_name}_{prop}_tuning.mol2")
    
    # Create CSV summary
    csv_filename = f"{molecule_name}_tuning_summary.csv"
    with open(csv_filename, 'w', newline='') as csvfile:
        fieldnames = ['point_index', 'x', 'y', 'z'] + [f'{prop}_effect' for prop in properties_to_calculate]
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        
        writer.writeheader()
        for i, (coord, effect) in enumerate(zip(surface_coords, all_effects)):
            row = {
                'point_index': i,
                'x': coord[0],
                'y': coord[1], 
                'z': coord[2]
            }
            # Add property effects
            for prop in properties_to_calculate:
                row[f'{prop}_effect'] = effect.get(f'{prop}_effect', 0.0)
            writer.writerow(row)
    
def check_all_files_created(molecule_name, surface_coords, properties_to_calculate):
    if len(surface_coords) == 1:
        core.print_office_quote()
    else:
        missing = []
        for prop in properties_to_calculate:
            filepath = f"{molecule_name}_{prop}.mol2"
            if not os.path.exists(filepath):
                missing.append(filepath)
        
        csv_path = f"{molecule_name}_tuning_summary.csv"
        if not os.path.exists(csv_path):
            missing.append(csv_path)

        if missing:
            print(f"Missing: {', '.join(missing)}")
        else:
            core.print_office_quote()


#################
# Miscellaneous #
#################

def startup_message():
    core.print_startup_message()
    print(f"="*60)
    print(f"                  Electrostatic Tuning Maps")
    print(f"             Built on efforts by the Gozem Lab")
    print(f"   See: https://pubs.acs.org/doi/10.1021/acs.jpcb.9b00489")
    print(f"="*60)
    print(f"\n")



# ####Main Execution####

#######################################
#           Preliminary Setup         #
#######################################

# Print startup message
startup_message()

####User Input####
# Get tuning file from command line argument if provided
tuning_file = sys.argv[1] if len(sys.argv) > 1 else 'tuning.in'

# Get parameters from tuning file
tuning_params = get_tuning_parameters(tuning_file)

#Input options
input_type = tuning_params.get('input_type', 'smiles')
input_data = tuning_params.get('input_data', 'O')

#Calculation options
method = tuning_params.get('method', 'dft')
basis_set = tuning_params.get('basis_set', 'cc-pVTZ')
functional = tuning_params.get('functional', 'b3lyp')
charge = tuning_params.get('charge', 0)
spin = tuning_params.get('spin', 0)
surface_charge = tuning_params.get('surface_charge', 1.0)
solvent = tuning_params.get('solvent', 'water')

#Calculation specifics
properties = tuning_params.get('properties', ['lumo'])
state_of_interest = tuning_params.get('state_of_interest', 2)
triplet = tuning_params.get('triplet', False)


# Check available hardware
No_of_GPUs = core.check_gpu_info()
No_of_CPUs = core.check_cpu_info()

# Resolve property dependencies and required calculations
properties_to_calculate, required_calculations = setup_calculation(properties)
print(f"Calculating Tuning of:  {properties_to_calculate}")
print(f"Using molecular states: {required_calculations}")


# Prepare input data
q_mm = np.array([surface_charge])
input_data = prepare_input_data(input_type, input_data)
molecule_name = core.extract_xyz_name(input_data)
surface_coords = core.get_vdw_surface_coordinates(input_data)
print(f"\n="*60)
print(f"   Running calculations on {len(surface_coords)} surface points")
print(f"="*60)
print(f"\n")


#######################################
#    Core Tuning Map Calculations     #
#######################################

# Create base molecule objects
base_molecules = create_molecule_objects(required_calculations, state_of_interest, triplet)

# Calculate effects at each surface point
all_effects = []
for i, coord in enumerate(surface_coords):
    effects = calculate_surface_effect_at_point(
        base_molecules, coord, surface_charge, 
        solvent, state_of_interest, triplet, properties_to_calculate, required_calculations
    )
    all_effects.append(effects)
    print(f"Point {i+1}/{len(surface_coords)}: {effects}")


# Create output files
create_output_files(surface_coords, all_effects, molecule_name, properties_to_calculate)
check_all_files_created(molecule_name, surface_coords, properties_to_calculate)



from pyscf import gto, dft, scf, lib
import time


lib.num_threads(8)

def create_charged_molecule_object(
    atom_input,
    basis_set,
    method='dft',
    functional='b3lyp',
    original_charge=0,
    charge_change=0,
    gpu=False,
    spin_guesses=None,
):
    charge = original_charge + charge_change
    spin_guesses = spin_guesses or [0, 1, 2, 3, 4]

    results = []  # store (spin, energy, mf)

    for spin in spin_guesses:
        try:
            mol = gto.Mole()
            mol.atom = atom_input
            mol.basis = basis_set
            mol.charge = charge
            mol.spin = spin   # number of unpaired electrons (2S)
            mol.build()

            # RKS for singlet, UKS for open shell
            if method.lower() == 'dft':
                mol.xc = functional
                mf = dft.UKS(mol) if spin > 0 else dft.RKS(mol)
            elif method.lower() == 'hf':
                mf = scf.UHF(mol) if spin > 0 else scf.RHF(mol)
            else:
                raise ValueError("Method must be 'dft' or 'hf'")
            if gpu:
                mf = mf.to_gpu()
            mf = mf.newton()
            energy = mf.kernel()
            if mf.converged:
                print(f"Spin {spin} (2S+1={spin+1}) converged: E = {energy:.6f} Ha")
                results.append((spin, energy, mf))
            else:
                print(f"Spin {spin} (2S+1={spin+1}) did NOT converge")

        except Exception as e:
            print(f"Spin {spin} failed: {e}")

    if results:
        # pick lowest energy among converged spins
        best_spin, best_energy, best_mf = min(results, key=lambda x: x[1])
        print(f"\nLowest energy: spin={best_spin} (2S+1={best_spin}), E={best_energy:.6f} Ha")
        return best_mf, best_spin, best_energy
    else:
        print("No spin converged for this species.")
        return None, None, None



best_mf, best_spin, best_energy = create_charged_molecule_object(
    atom_input='n-phe.xyz',
    basis_set="6-31g",
    method='dft',
    functional='b3lyp',
    original_charge=0,
    charge_change=-1,
    gpu=True,
    # spin_guesses=[0, 1]
)

print(best_spin, best_energy, best_mf.kernel() if best_mf else None)

# srun -p qGPU24 -A CHEM9C4 --nodes=1 --ntasks-per-node=1 --cpus-per-task=1 --gres=gpu:1 --time=01:00:00 --mem=8G --pty ems




import numpy as np
import os,subprocess,sys
from pyscf import gto,scf,dft,qmmm,tdscf
from pyscf.solvent import smd
from pyscf.hessian import thermo
from pyscf.geomopt.geometric_solver import optimize
from rdkit import Chem
from rdkit.Chem import AllChem


#Constants
HARTREE_TO_KCAL = 627.509
HARTREE_TO_EV = 27.2114
SHE_REFERENCE_POTENTIAL = 4.43

# Property dependency and unit mapping
PROPERTY_CONFIG = {
    'gse': {'deps': [], 'calc': [], 'unit': 1},
    'homo': {'deps': [], 'calc': [], 'unit': 1},
    'lumo': {'deps': [], 'calc': [], 'unit': 1},
    'gap': {'deps': ['homo', 'lumo'], 'calc': [], 'unit': 1},
    'dm': {'deps': [], 'calc': [], 'unit': 1},
    'ie': {'deps': [], 'calc': ['cation'], 'unit': HARTREE_TO_KCAL},
    'ea': {'deps': [], 'calc': ['anion'], 'unit': HARTREE_TO_KCAL},
    'cp': {'deps': ['ie', 'ea'], 'calc': [], 'unit': HARTREE_TO_KCAL},
    'eng': {'deps': ['cp'], 'calc': [], 'unit': HARTREE_TO_EV},
    'hard': {'deps': ['ie', 'ea'], 'calc': [], 'unit': HARTREE_TO_EV},
    'efl': {'deps': ['cp', 'hard'], 'calc': [], 'unit': HARTREE_TO_EV},
    'nfl': {'deps': ['efl'], 'calc': [], 'unit': HARTREE_TO_EV},
    'exe': {'deps': [], 'calc': ['td'], 'unit': 1},
    'osc': {'deps': [], 'calc': ['td'], 'unit': 1}
}

def setup_calculation(requested_props):
    """Setup properties and calculations needed"""
    if 'all' in requested_props:
        requested_props = list(PROPERTY_CONFIG.keys())
    
    # Resolve dependencies
    props_needed = set()
    def add_deps(prop):
        if prop in props_needed: return
        props_needed.add(prop)
        for dep in PROPERTY_CONFIG[prop]['deps']:
            add_deps(dep)
    
    for prop in requested_props:
        add_deps(prop)
    
    # Determine required calculations
    calcs_needed = {'neutral': True}
    for prop in props_needed:
        for calc in PROPERTY_CONFIG[prop]['calc']:
            calcs_needed[calc] = True
    
    return list(props_needed), calcs_needed

def calculate_all_properties(mf, anion_mf=None, cation_mf=None, td_obj=None, props_to_calc=None):
    results = {}
    
    # Basic properties
    if 'gse' in props_to_calc:
        results['gse'] = mf.e_tot * HARTREE_TO_KCAL
    
    if any(p in props_to_calc for p in ['homo', 'lumo', 'gap']):
        homo, lumo, gap = [x * HARTREE_TO_EV for x in find_homo_lumo_and_gap(mf)]
        results.update({p: v for p, v in zip(['homo', 'lumo', 'gap'], [homo, lumo, gap]) if p in props_to_calc})
    
    if 'dm' in props_to_calc:
        results['dm'] = np.linalg.norm(mf.dip_moment())
    
    # Charged state properties
    if 'ie' in props_to_calc and cation_mf:
        results['ie'] = cation_mf.e_tot - mf.e_tot
    if 'ea' in props_to_calc and anion_mf:
        results['ea'] = mf.e_tot - anion_mf.e_tot
    
    # Derived properties
    if 'cp' in props_to_calc and all(k in results for k in ['ie', 'ea']):
        results['cp'] = -(results['ie'] + results['ea']) / 2
    if 'eng' in props_to_calc and 'cp' in results:
        results['eng'] = -results['cp']
    if 'hard' in props_to_calc and all(k in results for k in ['ie', 'ea']):
        results['hard'] = (results['ie'] - results['ea']) / 2
    if 'efl' in props_to_calc and all(k in results for k in ['cp', 'hard']):
        results['efl'] = results['cp']**2 / (2 * results['hard']) if results['hard'] != 0 else 0
    if 'nfl' in props_to_calc and 'efl' in results:
        results['nfl'] = 1/results['efl'] if results['efl'] != 0 else 0
    
    # Excited state properties
    if td_obj and any(p in props_to_calc for p in ['exe', 'osc']):
        state_idx = state_of_interest - 1
        if 'exe' in props_to_calc:
            results['exe'] = td_obj.e[state_idx] * HARTREE_TO_EV
        if 'osc' in props_to_calc:
            results['osc'] = td_obj.oscillator_strength()[state_idx]
    
    return results

#Operations

#Molecular Object Creation
def create_molecule_object(atom_input, basis_set, method='dft', functional='m06-2x', charge=0, spin=1):
    """
    Create a PySCF molecule and KS/DFT object from either:
    - atom_input: list of (atom_symbol, (x,y,z)) tuples
    - atom_input: string path to XYZ file
    """
    spin -= 1
    mol = gto.M(basis = basis_set, charge=charge, spin=spin,symmetry=False)
    if isinstance(atom_input, str) and os.path.isfile(atom_input):
        mol.atom = atom_input 
    else:
        mol.atom = atom_input  
    mol.build()

    # Choose method and handle spin
    if method.lower() == 'dft':
        mol.xc=functional
        mf = dft.UKS(mol) if spin > 0 else dft.RKS(mol)
    elif method.lower() == 'hf':
        mf = scf.UHF(mol) if spin > 0 else scf.RHF(mol)
    else:
        raise ValueError("method must be 'dft' or 'hf'")
    return mf

def create_charged_molecule_object(atom_input, basis_set, method='dft', functional='m06-2x', original_charge=0, original_spin=1, charge_change=-1):
    """
    Create charged form of molecule (anion or cation) and return the lowest energy one with its spin and energy
    
    Args:
        charge_change: -1 for anion (add electron), +1 for cation (remove electron)
    """
    new_charge = original_charge + charge_change
    
    # Determine possible spin states based on charge change
    if charge_change == -1:  
        if original_spin == 2:  
            possible_spins = [1, 3] 
        elif original_spin == 1:  
            possible_spins = [2]  
        elif original_spin == 3:  
            possible_spins = [2, 4] 
            
    elif charge_change == +1:  
        if original_spin == 1:  
            possible_spins = [2]  
        elif original_spin == 2:  
            possible_spins = [1, 3]  
        elif original_spin == 3:  
            possible_spins = [2, 4] 
    
    energies = []
    objects = []
    
    for spin in possible_spins:
        pyscf_spin = spin - 1
        mol = gto.Mole(basis=basis_set, charge=new_charge, spin=pyscf_spin)
        mol.atom = atom_input
        mol.build()
        if method.lower() == 'dft':
            mf = dft.UKS(mol) if pyscf_spin > 0 else dft.RKS(mol)
            mf.xc = functional
        elif method.lower() == 'hf':
            mf = scf.UHF(mol) if pyscf_spin > 0 else scf.RHF(mol)
        energy = mf.kernel()
        energies.append(energy)
        objects.append(mf)
    
    # Find the lowest energy
    min_idx = energies.index(min(energies))
    lowest_energy = energies[min_idx]
    optimal_spin = possible_spins[min_idx]
    best_object = objects[min_idx]
    
    return best_object, optimal_spin, lowest_energy

def create_td_molecule_object(mf, nstates=5, triplet=False):
    # Handle solvated molecules
    if hasattr(mf, 'with_solvent'):
        # For solvated molecules, use tdscf directly
        if hasattr(mf, 'xc'):  # DFT case
            td = tdscf.TDDFT(mf)
        else:  # HF case
            td = tdscf.TDHF(mf)
    else:
        # For non-solvated molecules, use the original method
        if hasattr(mf, 'TDDFT') and mf.TDDFT is not None:
            td = mf.TDDFT()
        elif hasattr(mf, 'TDHF') and mf.TDHF is not None:
            td = mf.TDHF()
        else:
            raise ValueError("Unsupported ground state object type")
    
    td.singlet = not triplet
    td.nstates = nstates
    return td

#Manipulation and Extraction
def solvate_molecule(mf, solvent='water'):
    solvent = solvent.lower()
    mf = mf.PCM()
    mf.with_solvent.eps = smd.solvent_db[solvent][5]
    mf.with_solvent.method = 'C-PCM'
    mf.with_solvent.lebedev_order = 29
    return mf

def optimize_and_get_equilibrium(mf):
    """
    Optimize the geometry of the molecule and return the equilibrium geometry.
    """
    mol_eq = optimize(mf,conv_tol_grad=1e-7,conv_tol=1e-10)
    coords = mol_eq.atom_coords(unit='Ang')
    atoms = [mol_eq.atom_symbol(i) for i in range(mol_eq.natm)]
    atom_list = [(atom, coord) for atom, coord in zip(atoms, coords)]
    return atom_list

def smiles_to_xyz(smiles, filename=None):
    """Convert SMILES to XYZ file"""
    mol = Chem.AddHs(Chem.MolFromSmiles(smiles))
    AllChem.EmbedMolecule(mol)
    AllChem.MMFFOptimizeMolecule(mol)
    
    if not filename:
        filename = f"mol_{abs(hash(smiles)) % 10000}.xyz"
    
    conf = mol.GetConformer()
    with open(filename, 'w') as f:
        f.write(f"{mol.GetNumAtoms()}\n{smiles}\n")
        for i, atom in enumerate(mol.GetAtoms()):
            pos = conf.GetAtomPosition(i)
            f.write(f"{atom.GetSymbol()} {pos.x:.6f} {pos.y:.6f} {pos.z:.6f}\n")
    return filename

def create_optimized_molecule(atom_input, basis_set, method='dft', functional='m06-2x', charge=0, spin=1, optimize=True):
    """Create molecule with optional geometry optimization"""
    if optimize:
        # Create initial molecule for optimization
        mf_initial = create_molecule_object(atom_input, basis_set, method, functional, charge, spin)
        mf_initial.kernel()
        
        # Optimize geometry using existing function
        opt_coords = optimize_and_get_equilibrium(mf_initial)
        
        # Create new molecule with optimized coordinates
        return create_molecule_object(opt_coords, basis_set, method, functional, charge, spin)
    else:
        return create_molecule_object(atom_input, basis_set, method, functional, charge, spin)


def find_homo_lumo_and_gap(mf):
    homo = -float("inf")
    lumo = float("inf")
    for energy, occ in zip(mf.mo_energy, mf.mo_occ):
        if occ > 0 and energy > homo:
            homo = energy 
        if occ == 0 and energy < lumo:
            lumo = energy 
    return homo, lumo, lumo - homo

def get_molecule_gfec(mf):
    mf.kernel()
    hess = mf.Hessian().kernel()

    vib_data = thermo.harmonic_analysis(
        mf.mol,
        hess,
        imaginary_freq=False
    )

    thermo_data = thermo.thermo(
        mf,
        vib_data['freq_au'],
        temperature=298.15,
        pressure=101325
    )

    G_corr = thermo_data['G_tot'][0] - thermo_data['E0'][0]
    return vib_data, thermo_data, G_corr

#File Manipulation
def create_mol2_files(tuning, molecule_name, tuning_names):
    for property_name in tuning_names:
        filename = f"{molecule_name}_{property_name}_tm.mol2"
        property_data = tuning[property_name]
        num_points = len(property_data)
        
        with open(filename, 'w') as f:
            # Write header
            f.write("@<TRIPOS>MOLECULE\n")
            f.write(f"{filename}\n")
            f.write(f"     {num_points} 0 0 0\n")
            f.write("SMALL\n")
            f.write("GASTEIGER\n")
            f.write("@<TRIPOS>ATOM\n")
            
            # Write atom records
            for i, (coord, delta_value) in enumerate(property_data, 1):
                x, y, z = coord[0]  # Extract coordinates from (1,3) array
                f.write(f"{i:>4}    H  {x:>8.4f}  {y:>8.4f}  {z:>8.4f}  H1  1  {property_name.upper()}  {delta_value:>10.6f}\n")

#Parameters
molecule = 'water'
method = 'dft'
functional = 'b3lyp'
basis_set = 'augccpvdz'
charge = 0
spin = 1
gfec_functional = 'b3lyp'
gfec_basis_set = '6-31+G*'
state_of_interest = 2
triplet_excitation = False
solvent = None
rdx_solvent = 'acetonitrile'
input_type = 'xyz'  # 'xyz' or 'smiles'
smiles_input = 'O' 
optimize_geometry = True 

# User input and setup
# requested_properties = ['gse', 'homo', 'lumo', 'gap']  # User specifies
requested_properties = ['all']  # Or everything

properties_to_calculate, required_calculations = setup_calculation(requested_properties)
print(f"Properties: {properties_to_calculate}")
print(f"Calculations: {required_calculations}")

# Handle input and create molecule
if input_type == 'smiles':
    xyz_file = smiles_to_xyz(smiles_input)
    molecule_name = xyz_file.replace('.xyz', '')
else:
    xyz_file = f'{molecule}.xyz'
    molecule_name = molecule

molecule_object = create_molecule_object(xyz_file, basis_set, method=method, functional=functional, charge=charge, spin=spin)

if solvent:
    molecule_object = solvate_molecule(molecule_object, solvent=solvent)
    molecule_opt = optimize_and_get_equilibrium(molecule_object)
    molecule_object = create_molecule_object(molecule_opt, basis_set, method=method, functional=functional, charge=charge, spin=spin)
    molecule_object = solvate_molecule(molecule_object, solvent=solvent)

if optimize_geometry and solvent is None:
    molecule_object = create_optimized_molecule(molecule_object.mol.atom, basis_set, method=method, functional=functional, charge=charge, spin=spin, optimize=optimize_geometry)

print(molecule_object.mol.atom)

# Create XYZ file from molecule_object.mol.atom
atom_data = molecule_object.mol.atom
xyz_file = f"{molecule_name}_opt.xyz"
with open(xyz_file, 'w') as f:
    f.write(f"{len(atom_data)}\n")
    f.write(f"\n")
    for element, coords in atom_data:
        f.write(f"{element} {coords[0]:.6f} {coords[1]:.6f} {coords[2]:.6f}\n")
print(f"Created XYZ file: {xyz_file}")

molecule_object.kernel()

# Handle charged species with optional solvation
anion_mf = None
if required_calculations.get('anion'):
    anion_mf = create_charged_molecule_object(xyz_file, basis_set, method=method, functional=functional, original_charge=charge, original_spin=spin, charge_change=-1)[0]
    if solvent:
        anion_mf = solvate_molecule(anion_mf, solvent=solvent)
    anion_mf.kernel()

cation_mf = None
if required_calculations.get('cation'):
    cation_mf = create_charged_molecule_object(xyz_file, basis_set, method=method, functional=functional, original_charge=charge, original_spin=spin, charge_change=+1)[0]
    if solvent:
        cation_mf = solvate_molecule(cation_mf, solvent=solvent)
    cation_mf.kernel()

# Handle excited state calculations with optional solvation
td_object = None
if required_calculations.get('td'):
    td_object = create_td_molecule_object(molecule_object, triplet=triplet_excitation, nstates=state_of_interest)
    if solvent:
        td_object.with_solvent.equilibrium_solvation = True
    td_object.kernel()

# Calculate reference properties
properties_alone = calculate_all_properties(molecule_object, anion_mf, cation_mf, td_object, properties_to_calculate)

# Surface calculations
vdw_surface_operation = subprocess.run(['vsg', xyz_file, '--txt'], capture_output=True, text=True)
surface_coords = []
surface_file = xyz_file.replace('.xyz', '_vdw_surface.txt')
with open(surface_file, 'r') as f:
    for line in f:
        arr = np.fromstring(line.strip(), sep=' ').reshape(1, 3)
        surface_coords.append(arr)

tuning = {name: [] for name in properties_to_calculate}

for coord in surface_coords:
    # Create perturbed molecules
    molecule_wsc = qmmm.mm_charge(molecule_object, coord, np.array([1.0]))
    molecule_wsc.kernel()
    
    anion_wsc = qmmm.mm_charge(anion_mf, coord, np.array([1.0])) if anion_mf else None
    if anion_wsc: anion_wsc.kernel()
    
    cation_wsc = qmmm.mm_charge(cation_mf, coord, np.array([1.0])) if cation_mf else None
    if cation_wsc: cation_wsc.kernel()
    
    td_wsc = None
    if td_object:
        td_wsc = create_td_molecule_object(molecule_wsc, triplet=triplet_excitation, nstates=state_of_interest)
        if solvent:
            td_wsc.with_solvent.equilibrium_solvation = True
        td_wsc.kernel()
    
    # Calculate perturbed properties and store deltas
    properties_wsc = calculate_all_properties(molecule_wsc, anion_wsc, cation_wsc, td_wsc, properties_to_calculate)
    
    for prop in properties_to_calculate:
        if prop in properties_wsc and prop in properties_alone:
            delta = (properties_wsc[prop] - properties_alone[prop]) * PROPERTY_CONFIG[prop]['unit']
            tuning[prop].append([coord, delta])

create_mol2_files(tuning, molecule_name, properties_to_calculate)
print(f"Created {len(properties_to_calculate)} MOL2 files for {molecule_name}")

